import os
import shutil
from pdb import pm
from miasm2.analysis.sandbox import Sandbox_Linux_x86_32
from miasm2.core.asmbloc import bbl_simplifier, expr_is_label, asm_resolve_final
from miasm2.expression.expression import ExprId
from miasm2.core.interval import interval
from miasm2.ir.symbexec import symbexec


# Parse arguments
parser = Sandbox_Linux_x86_32.parser(description="ELF sandboxer")
parser.add_argument("filename", help="ELF Filename")
options = parser.parse_args()

# Create sandbox
sb = Sandbox_Linux_x86_32(options.filename, options, globals())

machine = sb.machine
bin_stream = sb.jitter.bs
entry_point = sb.entry_point

mdis = machine.dis_engine(bin_stream)

to_exec = {}
def cb_decrypt(cur_bloc, offsets_to_dis, *args, **kwargs):
    start_addr = to_exec.get(cur_bloc.label, None)
    if start_addr:
        # Pattern detected, exec it
        print "Match on", hex(start_addr)
        # Reset exception, in case of disassembly error
        sb.jitter.vm.set_exception(0)
        # Place a breakpoint at the end of the pattern
        bp_addr = cur_bloc.lines[-1].offset
        sb.jitter.add_breakpoint(bp_addr, lambda x: False)
        # Run the emulation from the head
        sb.run(start_addr)
        # Sanity check
        assert sb.jitter.pc == bp_addr

    if cur_bloc.lines and cur_bloc.lines[-1].name == "INT":
        # In case of exit interuption, do not continue to disasm
        ## Symbol exec the current block to get EAX value
        ira = machine.ira()
        ira.add_bloc(cur_bloc)
        symb = symbexec(ira,
                        ira.arch.regs.regs_init)
        symb.emul_ir_bloc(ira, cur_bloc.label.offset)
        intnb = symb.symbols[ira.arch.regs.EAX]
        if int(intnb.arg) == 1:
            # sys_exit, do not disassemble successors
            cur_bloc.bto.clear()
            offsets_to_dis.clear()

    # Detect pattern head
    if not (len(cur_bloc.lines) >= 5 and
            cur_bloc.lines[0].name == "PUSH" and
            cur_bloc.lines[1].name == "MOV" and
            cur_bloc.lines[-1].name.startswith("J") and
            cur_bloc.lines[-1].name != "JMP"):
        return

    # Detection of the pattern, remember the start addr
    to_exec[cur_bloc.get_next()] = cur_bloc.label.offset

mdis.dis_bloc_callback = cb_decrypt
blocks = mdis.dis_multibloc(entry_point)


from miasm2.core.graph import MatchGraphJoker
# Pattern we want to match:
#           |
#      +----v----+
#      |  (dad)  |
#      |  PUSH   |
#      |  MOV    |
#      +----+----+
#           |
#      +----v----+
#      | (middle)|<---+
#      +----+--+-+    |
#           |  +------+
#      +----v----+
#      |  (end)  |
#      +----+----+
#           |
#           v

# First block must have 2 assembly lines, a PUSH then a MOV, and an unlimited
# number of incomming edge
dad = MatchGraphJoker(name="dad", restrict_in=False,
                      filt=lambda block: (len(block.lines) == 2 and
                                          block.lines[0].name == "PUSH" and
                                          block.lines[1].name == "MOV"))
middle = MatchGraphJoker(name="middle")
# Last block can have an unlimited number of outcomming edge
end = MatchGraphJoker(name="end", restrict_out=False)
# Build a MatchGraph by specifing edges (A >> B means an edge from A to B)
matcher = dad >> middle >> middle >> end

def block_merge(dgs, graph):
    """Remove the pattern defined in matcher
    @dgs: instance of DiGraphSimplifier (for recursive call purpose)
    @graph: graph to simplify
    """
    for sol in matcher.match(graph):
        # /!\ we modify the graph during an iteration on it
        # Our match are disjoint, we can avoid breaking the loop on each match
        # Links each successors to our pattern predecessors
        successors = graph.successors(sol[end])
        for pred in graph.predecessors(sol[dad]):
            for succ in successors:
                constraint = graph.edges2constraint[(pred, sol[dad])]
                # Patch jump
                last_instr = pred.lines[-1]
                if last_instr.breakflow() and last_instr.dstflow():
                    ## Expr are immutable, so build a new arg list
                    new_arg = []
                    for arg in last_instr.args:
                        if expr_is_label(arg):
                            # Link pred -> succ, the new destination
                            new_arg.append(ExprId(succ.label, arg.size))
                        else:
                            new_arg.append(arg)
                    last_instr.args = new_arg
                # Update links
                graph.add_edge(pred, succ, constraint)

        # Remove matched blocks (and associated edges) from the graph
        for node in sol.itervalues():
            graph.del_node(node)

from miasm2.core.graph import DiGraphSimplifier
dgs = DiGraphSimplifier()
dgs.enable_passes([block_merge])

blocks_after = dgs(blocks)
blocks_final = bbl_simplifier(blocks_after)
open("cfg_after.dot", "w").write(blocks_final.dot())
blocks_final.sanity_check()

head = blocks_final.heads()[0]
for label in mdis.symbol_pool.items:
    mdis.symbol_pool.del_label_offset(label)
mdis.symbol_pool.set_offset(head.label, entry_point)

dinterval = interval(blk.get_range() for blk in blocks)
dinterval_full = interval([dinterval.hull()])
patches = asm_resolve_final(mdis.arch, blocks_final, mdis.symbol_pool,
                            dst_interval=dinterval_full)
new_filename = "%s_clean.bin" % options.filename
shutil.copy(options.filename, new_filename)

# Use Elfesteem to translate the address to the corresponding offset
section_ep = sb.elf.getsectionbyvad(entry_point).sh
offset_start = section_ep.addr - section_ep.offset

with open(new_filename, "r+") as fdesc:
    # Clean older values
    for start, stop in dinterval.intervals:
        ## Calc offset
        fdesc.seek(start - offset_start)
        fdesc.write("\x00" * (stop - start + 1))

    # Write new values
    for offset, data in patches.iteritems():
        fdesc.seek(offset - offset_start)
        fdesc.write(data)

